@@ -9,6 +9,7 @@ use std::ptr::NonNull;
 
 type Link<K, V> = Option<NonNull<SkipNode<K, V>>>;
 
+// A node in the skip list.
 #[derive(Debug)]
 struct SkipNode<K: Ord + Hash + Debug, V: Clone> {
     /// The key should only be `None` for the `head` node.
@@ -34,6 +35,30 @@ impl<K: Ord + Hash + Debug, V: Clone> SkipNode<K, V> {
             levels: vec![],
         }
     }
+
+    /// Get an immutable reference to the next node at level 0 if it exists. Otherwise, `None`.
+    fn next(&self) -> Option<&SkipNode<K, V>> {
+        self.next_at_level(0)
+    }
+
+    /// Get an immutable reference to the next node at the specified level if it exists.
+    /// Otherwise, `None`.
+    fn next_at_level(&self, level: usize) -> Option<&SkipNode<K, V>> {
+        if self.levels.is_empty() {
+            return None;
+        }
+
+        self.levels[level].as_ref().map(|node_ptr| {
+            unsafe {
+                /*
+                SAFETY:
+                This is safe because links are guaranteed to exist. If the link did not exist, it
+                would be a `None` in the tower and execution would not have gotten here.
+                */
+                node_ptr.as_ref()
+            }
+        })
+    }
 }
 
 /// A skip list.
@@ -92,45 +117,8 @@ impl<K: Ord + Hash + Debug, V: Clone> SkipList<K, V> {
             return None;
         }
 
-        let mut current_node = &*self.head;
-
-        // Start iteration at the top of the skip list "towers" and iterate through pointers at the
-        // current level. If we skipped past our key, move down a level.
-        for level_idx in (0..self.height()).rev() {
-            // Get an optional of the next node
-            let mut maybe_next_node = current_node.levels[level_idx].as_ref();
-
-            while maybe_next_node.is_some() {
-                let next_node_ptr = maybe_next_node.unwrap();
-                let next_node = unsafe {
-                    // SAFETY: next_node_ptr is guaranteed to exist by the condition for the `while`
-                    next_node_ptr.as_ref()
-                };
-                match next_node.key.as_ref().unwrap().cmp(key) {
-                    std::cmp::Ordering::Less => {
-                        current_node = next_node;
-                        maybe_next_node = next_node.levels[level_idx].as_ref();
-                    }
-                    _ => break,
-                }
-            }
-        }
-
-        // The while loop uses a less than comparator and stops at the node that is potentially just
-        // prior to the node we are looking for. We need to move the pointer forward one time and
-        // check we actually arrived at our node or if we hit the end of the levels without finding
-        // anything.
-        let potential_record = current_node.levels[0].as_ref();
-        if potential_record.is_some() {
-            let record = unsafe {
-                /*
-                SAFETY:
-                The `is_some` check ensures that the pointer exists and globally all links are
-                guaranteed to be valid.
-                */
-                potential_record.as_ref().unwrap().as_ref()
-            };
-
+        let potential_record = self.find_greater_or_equal_node(key);
+        if let Some(record) = potential_record {
             if record.key.as_ref().unwrap().eq(key) {
                 return record.value.as_ref();
             }
@@ -320,7 +308,7 @@ impl<K: Ord + Hash + Debug, V: Clone> SkipList<K, V> {
             SAFETY:
             The other references to this pointer were removed when removing the links above, so
             `found_node_ptr` should be the last reference to this node. All links should be valid so
-            `found_ndde_ptr`, which is a link at level 0, should be valid.
+            `found_node_ptr`, which is a link at level 0, should be valid.
             */
             Box::from_raw(found_node_ptr.as_ptr())
         };
@@ -343,7 +331,9 @@ impl<K: Ord + Hash + Debug, V: Clone> SkipList<K, V> {
         self.approximate_mem_usage
     }
 
-    /// An iterator visiting each node.
+    /// An iterator visiting each node in order.
+    ///
+    /// Returns values of (&'a K, &'a V)
     pub fn iter(&self) -> NodeIterHelper<'_, K, V> {
         if self.is_empty() {
             return NodeIterHelper { next: None };
@@ -356,6 +346,76 @@ impl<K: Ord + Hash + Debug, V: Clone> SkipList<K, V> {
 
         NodeIterHelper { next }
     }
+
+    /// An iterator visiting each node in order with mutable references to values.
+    ///
+    /// Returns values of (&'a K, &'a mut V)
+    pub fn iter_mut(&mut self) -> NodeIterMutHelper<'_, K, V> {
+        if self.is_empty() {
+            return NodeIterMutHelper { next: None };
+        }
+
+        let next = self.head.levels[0].as_mut().map(|node_ptr| unsafe {
+            // SAFETY: All links are valid if they exist.
+            node_ptr.as_mut()
+        });
+
+        NodeIterMutHelper { next }
+    }
+
+    /// Return a reference to the key and value of the first node in the skip list if there is a
+    /// node. Otherwise, it returns `None`.
+    pub fn first(&self) -> Option<(&K, &V)> {
+        self.head
+            .next()
+            .map(|node| (node.key.as_ref().unwrap(), node.value.as_ref().unwrap()))
+    }
+
+    /// Return a reference to the key and value of the last node in the skip list if there is a
+    /// node. Otherwise, it returns `None`.
+    pub fn last(&self) -> Option<(&K, &V)> {
+        if self.is_empty() {
+            return None;
+        }
+
+        let mut current_node = &*self.head;
+        for level_idx in (0..self.height()).rev() {
+            let mut maybe_next_node = current_node.next_at_level(level_idx);
+            while let Some(next_node) = maybe_next_node {
+                current_node = next_node;
+                maybe_next_node = next_node.next_at_level(level_idx);
+            }
+        }
+
+        Some((
+            current_node.key.as_ref().unwrap(),
+            current_node.value.as_ref().unwrap(),
+        ))
+    }
+
+    /// Return a reference to the key and value of the first node with a key that is greater than
+    /// or equal to the target key.
+    pub fn find_greater_or_equal(&self, target: &K) -> Option<(&K, &V)> {
+        if self.is_empty() {
+            return None;
+        }
+
+        self.find_greater_or_equal_node(target)
+            .map(|node| (node.key.as_ref().unwrap(), node.value.as_ref().unwrap()))
+    }
+
+    /// Return a reference to the key and value of the last node with a key that is less than the
+    /// target key.
+    pub fn find_less_than(&self, target: &K) -> Option<(&K, &V)> {
+        self.find_less_than_node(target).and_then(|node| {
+            // Only the head node has empty keys and values, so this means the search
+            // stayed on the head node. This can happen if we are searching for a target
+            // with less than all of the keys in the skip list. Return `None` in this case.
+            node.key.as_ref()?;
+
+            Some((node.key.as_ref().unwrap(), node.value.as_ref().unwrap()))
+        })
+    }
 }
 
 /// Implementation for keys and values that implement `Clone`
@@ -449,6 +509,73 @@ impl<K: Ord + Hash + Debug, V: Clone> SkipList<K, V> {
     fn dec_length(&mut self) {
         self.length -= 1;
     }
+
+    /// Return a reference to the first node with a key that is greater than or equal to the target
+    /// key.
+    fn find_greater_or_equal_node(&self, target: &K) -> Option<&SkipNode<K, V>> {
+        if self.is_empty() {
+            return None;
+        }
+
+        let mut current_node = &*self.head;
+        // Start iteration at the top of the skip list "towers" and iterate through pointers at the
+        // current level. If we skipped past our key, move down a level.
+        for level_idx in (0..self.height()).rev() {
+            // Get an optional of the next node
+            let mut maybe_next_node = current_node.next_at_level(level_idx);
+
+            while let Some(next_node) = maybe_next_node {
+                match next_node.key.as_ref().unwrap().cmp(target) {
+                    std::cmp::Ordering::Less => {
+                        current_node = next_node;
+                        maybe_next_node = next_node.next_at_level(level_idx);
+                    }
+                    std::cmp::Ordering::Equal | std::cmp::Ordering::Greater => {
+                        if level_idx == 0 {
+                            // We are at the bottom of the tower, so this is closest node greater
+                            // than or equal to our target.
+                            return Some(next_node);
+                        }
+
+                        // We found a node greater than or equal to our target. See if this is the
+                        // the first greatest node after our target by breaking and moving one
+                        // level down in the tower.
+                        break;
+                    }
+                }
+            }
+        }
+
+        // This is reached when the target is greater than all of the nodes in the skip list.
+        None
+    }
+
+    /// Return a reference to the last node with a key that is less than the target key.
+    fn find_less_than_node(&self, target: &K) -> Option<&SkipNode<K, V>> {
+        if self.is_empty() {
+            return None;
+        }
+
+        let mut current_node = &*self.head;
+        // Start iteration at the top of the skip list "towers" and iterate through pointers at the
+        // current level. If we skipped past our key, move down a level.
+        for level_idx in (0..self.height()).rev() {
+            // Get an optional of the next node
+            let mut maybe_next_node = current_node.next_at_level(level_idx);
+
+            while let Some(next_node) = maybe_next_node {
+                match next_node.key.as_ref().unwrap().cmp(target) {
+                    std::cmp::Ordering::Less => {
+                        current_node = next_node;
+                        maybe_next_node = next_node.next_at_level(level_idx);
+                    }
+                    _ => break,
+                }
+            }
+        }
+
+        Some(current_node)
+    }
 }
 
 /// An iterator adapter over the nodes of a `SkipList`.
@@ -516,6 +643,82 @@ where
 {
 }
 
+/// An mutable iterator adapter over the nodes of a `SkipList`.
+///
+/// This `struct` is created by the [`iter_mut`] method.
+///
+/// [`iter_mut`]: SkipList::iter_mut
+pub struct NodeIterMutHelper<'a, K, V>
+where
+    K: Ord + Hash + Debug,
+    V: Clone,
+{
+    next: Option<&'a mut SkipNode<K, V>>,
+}
+
+impl<'a, K, V> Iterator for NodeIterMutHelper<'a, K, V>
+where
+    K: Ord + Hash + Debug,
+    V: Clone,
+{
+    type Item = (&'a K, &'a mut V);
+
+    fn next(&mut self) -> Option<Self::Item> {
+        match self.next.take() {
+            None => None,
+            Some(current_node) => {
+                self.next = unsafe {
+                    /*
+                    SAFETY:
+                    Links at level 0 are always valid. No mutations can happen because the only way to get a
+                    `NodeIterator` is via [`SkipList::iter`] which borrows an immutable reference.
+                    */
+                    current_node.levels[0]
+                        .as_mut()
+                        .map(|node_ptr| node_ptr.as_mut())
+                };
+
+                return Some((
+                    current_node.key.as_ref().unwrap(),
+                    current_node.value.as_mut().unwrap(),
+                ));
+            }
+        }
+    }
+}
+
+impl<K, V> Drop for SkipList<K, V>
+where
+    K: Ord + Hash + Debug,
+    V: Clone,
+{
+    fn drop(&mut self) {
+        if self.is_empty() {
+            return;
+        }
+
+        let mut maybe_node_ptr = self.head.as_mut().levels[0];
+
+        while maybe_node_ptr.is_some() {
+            let mut current_node_ptr = maybe_node_ptr.unwrap();
+
+            /*
+            Re-box the allocation the pointer represents so that it can get dropped. Insert's
+            leak the boxed node when it is created.
+
+            It is ok to leave pointers in the dropped node's levels vector dangling because all
+            nodes are getting dropped.
+            */
+            let current_node = unsafe {
+                // SAFETY: All links are guaranteed to be valid nodes.
+                Box::from_raw(current_node_ptr.as_mut())
+            };
+
+            maybe_node_ptr = current_node.levels[0];
+        }
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -788,4 +991,126 @@ mod tests {
         usage_approximation -= base_node_usage + 8 + skiplist.height() * link_size;
         assert!(skiplist.get_approx_mem_usage() >= usage_approximation);
     }
+
+    #[test]
+    fn with_a_non_empty_skiplist_first_returns_references_to_the_first_key_value_pair() {
+        let mut skiplist = SkipList::<i32, String>::new(None);
+        skiplist.insert(2, "banana".to_string());
+        skiplist.insert(3, "orange".to_string());
+        skiplist.insert(1, "apple".to_string());
+        skiplist.insert(4, "strawberry".to_string());
+        skiplist.insert(5, "watermelon".to_string());
+
+        assert_eq!(skiplist.first(), Some((&1, &"apple".to_string())));
+    }
+
+    #[test]
+    fn with_an_empty_skiplist_first_returns_none() {
+        let skiplist = SkipList::<i32, String>::new(None);
+
+        assert_eq!(skiplist.first(), None);
+    }
+
+    #[test]
+    fn with_a_non_empty_skiplist_last_returns_references_to_the_last_key_value_pair() {
+        let mut skiplist = SkipList::<i32, String>::new(None);
+        skiplist.insert(2, "banana".to_string());
+        skiplist.insert(3, "orange".to_string());
+        skiplist.insert(1, "apple".to_string());
+        skiplist.insert(4, "strawberry".to_string());
+        skiplist.insert(5, "watermelon".to_string());
+
+        assert_eq!(skiplist.last(), Some((&5, &"watermelon".to_string())));
+    }
+
+    #[test]
+    fn with_an_empty_skiplist_last_returns_none() {
+        let skiplist = SkipList::<i32, String>::new(None);
+
+        assert_eq!(skiplist.last(), None);
+    }
+
+    #[test]
+    fn with_a_non_empty_skiplist_find_greater_or_equal_returns_correct_responses() {
+        let mut skiplist = SkipList::<i32, String>::new(None);
+        skiplist.insert(2, "banana".to_string());
+        skiplist.insert(3, "orange".to_string());
+        skiplist.insert(1, "apple".to_string());
+        skiplist.insert(4, "strawberry".to_string());
+        skiplist.insert(5, "watermelon".to_string());
+        skiplist.insert(11, "grapefruit".to_string());
+        skiplist.insert(12, "mango".to_string());
+
+        assert_eq!(
+            skiplist.find_greater_or_equal(&1),
+            Some((&1, &"apple".to_string())),
+            "The target is the first element so it should be found"
+        );
+
+        assert_eq!(
+            skiplist.find_greater_or_equal(&3),
+            Some((&3, &"orange".to_string())),
+            "The middle element exists so it should be found"
+        );
+
+        assert_eq!(
+            skiplist.find_greater_or_equal(&7),
+            Some((&11, &"grapefruit".to_string())),
+            "The target does not exist but there is a greater node so it should return that node"
+        );
+
+        assert_eq!(
+            skiplist.find_greater_or_equal(&12),
+            Some((&12, &"mango".to_string())),
+            "THe last element exists so it should be found"
+        );
+
+        assert_eq!(
+            skiplist.find_greater_or_equal(&20),
+            None,
+            "The target is greater than every element in the list so `None` should be returned"
+        );
+    }
+
+    #[test]
+    fn with_a_non_empty_skiplist_find_less_than_returns_correct_responses() {
+        let mut skiplist = SkipList::<i32, String>::new(None);
+        skiplist.insert(2, "banana".to_string());
+        skiplist.insert(3, "orange".to_string());
+        skiplist.insert(1, "apple".to_string());
+        skiplist.insert(4, "strawberry".to_string());
+        skiplist.insert(5, "watermelon".to_string());
+        skiplist.insert(11, "grapefruit".to_string());
+        skiplist.insert(12, "mango".to_string());
+
+        assert_eq!(
+            skiplist.find_less_than(&0),
+            None,
+            "Finding a target less than every element in the list returns `None`"
+        );
+
+        assert_eq!(
+            skiplist.find_less_than(&1),
+            None,
+            "Finding a target less than the first element returns `None`"
+        );
+
+        assert_eq!(
+            skiplist.find_less_than(&3),
+            Some((&2, &"banana".to_string())),
+            "Finding a target less than an existing middle element"
+        );
+
+        assert_eq!(
+            skiplist.find_less_than(&7),
+            Some((&5, &"watermelon".to_string())),
+            "Finding a target less than a non-existent middle element"
+        );
+
+        assert_eq!(
+            skiplist.find_less_than(&20),
+            Some((&12, &"mango".to_string())),
+            "Finding a target greater than all elements returns the last element"
+        );
+    }
 }
